import dataclasses
import logging

import jinja2

from . import common

_logger = logging.getLogger(__name__)


data_dir = common.data_dir
send_pull_pairs = []


@dataclasses.dataclass
class ManifestContainer:
    manifest: str
    template: jinja2.Template


@dataclasses.dataclass
class SendPullPair:
    send: ManifestContainer
    pull: ManifestContainer


send = ManifestContainer(
    manifest="sendstream.yaml",
    template=common.env.get_template("sendstream.yaml.j2"),
)

pull = ManifestContainer(
    manifest="pulltest.yaml", template=common.env.get_template("pulltest.yaml.j2")
)

pair = SendPullPair(send, pull)

send_pull_pairs.append(pair)


send = ManifestContainer(
    manifest="sendstream2.yaml",
    template=common.env.get_template("sendstream2.yaml.j2"),
)

pull = ManifestContainer(
    manifest="pulltest2.yaml", template=common.env.get_template("pulltest2.yaml.j2")
)

pair = SendPullPair(send, pull)

send_pull_pairs.append(pair)


def delete_old_manifests():
    _logger.debug(f"{data_dir=}")


def generate_manifest(session, container, directory):
    data = common.generate_data(session)

    x = session.encoder.lower()
    y = session.decoder.lower()
    pair_annotation = f"{x}-{y}".replace("$", "")

    x = f"{pair_annotation}-{container.manifest}"

    x = x.replace("$", "")

    path = directory / x
    text = container.template.render(data=data)
    _logger.info(f"writing {path.resolve()}")

    path.parent.mkdir(exist_ok=True, parents=True)
    path.write_text(text)


def generate_k8s(session):
    delete_old_manifests()

    for index, pair in enumerate(send_pull_pairs, 1):
        x = f"k8s-{str(session.encoder).replace('$', '')}-{index}"
        _dir = data_dir / x
        generate_manifest(session, pair.send, directory=_dir)
        generate_manifest(session, pair.pull, directory=_dir)



import collections
import io
import logging
import os
import pathlib
import pprint
import sys

import jinja2
import peewee
import requests

from . import common

_logger = logging.getLogger(__name__)

outfile = pathlib.Path("geo.sh")
request_url = "https://api.ip2location.io"
session = requests.Session()
db_path = pathlib.Path("geo.db")
db = peewee.SqliteDatabase(db_path)


class IP2L(peewee.Model):
    ip = peewee.CharField()
    country_code = peewee.CharField(null=True)
    country_name = peewee.CharField(null=True)
    region_name = peewee.CharField(null=True)
    city_name = peewee.CharField(null=True)
    latitude = peewee.CharField(null=True)
    longitude = peewee.CharField(null=True)
    zip_code = peewee.CharField(null=True)
    time_zone = peewee.CharField(null=True)
    asn = peewee.CharField(null=True)
    as_ = peewee.CharField(null=True)
    is_proxy = peewee.CharField(null=True)

    class Meta:
        database = db  # see module level db var
        db_table = "geo"


def view2():
    q_records = IP2L.select().dicts()
    records = {}
    for dct in q_records:
        region = dct["region_name"]
        ip = dct["ip"]
        if region == "-":
            region = "internal"
        records.setdefault(region, []).append(ip)

    loader = jinja2.FileSystemLoader(searchpath=common.templates_dir)
    env = jinja2.Environment(loader=loader, keep_trailing_newline=True)
    template = env.get_template("geo_view2.j2")
    text = template.render(data=records)
    print(text)


def view1():
    q_records = IP2L.select()
    q_regions = IP2L.select().group_by(IP2L.region_name)

    regions = [x.region_name for x in q_regions]
    c = collections.Counter(regions)
    for rec in q_records:
        c[rec.region_name] += 1
    print(c)


def ip_geolocation(ips: list[str]) -> dict:
    def quick_check(ips):
        vfile = io.StringIO("")

        vfile.write("#!/usr/bin/env bash")
        vfile.write("\n")

        for ip in ips:
            out = f"""
            curl 'https://api.ip2location.io/?key={key}&ip={ip}'
            """
            out = out.strip()
            vfile.write(out)

            vfile.write("\n")
            vfile.write("echo")

            vfile.write("\n")
        return vfile

    key = get_api_key_from_env()

    _logger.debug("quick check")
    strio = quick_check(ips).getvalue()
    debug = f"\n{strio}"

    outfile.write_text(debug)

    _logger.debug(debug)


def get_api_key_from_env() -> str:
    key = os.getenv("VAR_IP2LOCATION_API_KEY", None)

    if not key:
        raise ValueError("VAR_IP2LOCATION_API_KEY")

    return key


def fetch_data_for_ip(ip: str, api_key: str) -> dict:
    payload = {
        "key": api_key,
        "ip": ip,
    }

    try:
        response = requests.get(request_url, params=payload, timeout=5)
        _logger.debug(f"{response.url=}")

        # Raises a HTTPError if the status is 4xx, 5xxx
        response.raise_for_status()

    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        _logger.critical(f"can't reach {request_url}")
        sys.exit(-1)

    except requests.exceptions.HTTPError:
        _logger.critical("4xx, 5xx")
        raise

    else:
        _logger.debug(f"Great!  I'm able to reach {request_url}")

    _logger.debug(response.text)

    return response.json()


def db_initialize():
    if not db_path.exists():
        _logger.info(f"creating db {db_path.resolve()}")
    IP2L.create_table()


def get_regions_for_ips(ips: list[str]) -> None:
    db_initialize()
    api_key = get_api_key_from_env()

    for ip in ips:
        records = IP2L.select().where(IP2L.ip == ip).dicts()
        _logger.debug(f"fetching info for {ip}")

        for row in records:
            pf = pprint.pformat(row)
            _logger.debug(f"found record for ip {ip} in local cache")
            _logger.debug(pf)

        if not records:
            msg = (
                f"{ip} not found locally, reaching out "
                f"to {request_url} to fetch info"
            )
            _logger.debug(msg)
            dct = fetch_data_for_ip(ip, api_key)
            dct["as_"] = dct["as"]
            del dct["as"]
            ipl = IP2L(**dct)
            _logger.debug(f"saving {dct=} to cache")
            ipl.save()


def test_fetch_and_save():
    db_initialize()
    api_key = get_api_key_from_env()
    ips = [
        "172.30.1.236",
        "89.189.176.193",
        "98.97.56.92",
    ]

    for ip in ips:
        records = IP2L.select().where(IP2L.ip == ip).dicts()
        for row in records:
            pprint.pprint(row)

        if not records:
            dct = fetch_data_for_ip(ip, api_key)
            dct["as_"] = dct["as"]
            del dct["as"]
            pprint.pprint(dct)
            ipl = IP2L(**dct)
            ipl.save()


if __name__ == "__main__":
    view2()
#    view1()
# test_fetch_and_save()



import dataclasses
import datetime


@dataclasses.dataclass
class Credentials:
    login: str
    password: str


@dataclasses.dataclass
class LightSession:
    encoder: str
    decoder: str
    port: int
    expires_at: datetime.datetime

    def __post_init__(self):
        if not self.encoder:
            raise ValueError("encoder session id should not be empty")

        if not self.decoder:
            raise ValueError("decoder session id should not be empty")



import logging
import os
import pathlib

import jinja2
import pkg_resources

global delay_to_prevent_crash
delay_to_prevent_crash = None


_logger = logging.getLogger(__name__)

data_dir = pathlib.Path("data")
_logger.info(f"data directory is {data_dir.resolve()}")
data_dir.mkdir(parents=True, exist_ok=True)


package = __name__.split(".")[0]
templates_dir = pathlib.Path(pkg_resources.resource_filename(package, "templates"))
loader = jinja2.FileSystemLoader(searchpath=templates_dir)
env = jinja2.Environment(loader=loader, keep_trailing_newline=True)

msg_base_url_invalid = (
    "VAR_SLS_LIGHT_BASE_URL is required."
    " Example: http://hostname.domainname.com/, "
    "quitting prematurely."
)

base_url = os.getenv("VAR_SLS_LIGHT_BASE_URL", None)
status_url = f"{base_url}/light/light_status.php"
request_url = f"{base_url}/light/sreq.php"


def generate_pulltest_login():
    pulltest_login = os.getenv("VAR_PULLTEST_LOGIN", None)

    if not pulltest_login:
        raise ValueError("VAR_PULLTEST_LOGIN not defined")

    return pulltest_login


def generate_pulltest_password():
    pulltest_password = os.getenv("VAR_PULLTEST_PASSWORD", None)

    if not pulltest_password:
        raise ValueError("VAR_PULLTEST_PASSWORD not defined")

    return pulltest_password


def generate_data(session):
    pulltest_password = generate_pulltest_password()
    pulltest_login = generate_pulltest_login()

    encoder_decoder_id = f"{session.encoder}-{session.decoder}"
    slug = f"slug-{encoder_decoder_id}".lower().replace("$", "")
    reporter = f"reporter-{encoder_decoder_id}".lower().replace("$", "")

    data = {
        "ip": "172.30.0.139",
        "encoder_decoder_id": encoder_decoder_id,
        "drm": session.encoder,
        "network1": session.encoder,
        "NET1": session.decoder,
        "reporter": reporter,
        "slug": slug,
        "decoder": session.decoder,
        "pull_port": session.port,
        "pulltest_login": pulltest_login,
        "pulltest_password": pulltest_password,
    }
    return data



import datetime
import logging
import os
import pprint
import random
import re
import sys
import time
import typing

import bs4
import durations
import pandas
import pytz
import requests
import requests.exceptions

from . import common, geolocate, k8s, model, scripts

_logger = logging.getLogger(__name__)

session_pair_pat = re.compile(
    r"Session pair has been generated\(dec\)"
    r" (?P<dec>\$[^:]+)\s+:\s+(?P<enc>\$[^:]+) \(enc\)"
)

sls_offline_pat = re.compile(r"ERROR: SLS service is offline")

CONNECT_TIMEOUT_SEC = 2

if not common.base_url:
    raise ValueError(common.msg_base_url_invalid)

# use requests's session auto manage cookies
module_session = requests.Session()

host_timezone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo


def avoid_sls_crash():
    time.sleep(common.delay_to_prevent_crash.total_seconds())


def check_host_is_running(endpoint: str) -> None:
    try:
        _logger.debug(f"feching {endpoint}")
        response = requests.get(endpoint, timeout=CONNECT_TIMEOUT_SEC)

        # Raises a HTTPError if the status is 4xx, 5xxx
        response.raise_for_status()

    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        _logger.critical(f"can't reach {endpoint}")
        sys.exit(-1)

    except requests.exceptions.HTTPError:
        _logger.critical("4xx, 5xx")
        sys.exit(-1)

    else:
        _logger.debug(f"Great!  I'm able to reach {endpoint}")


def populate_login_session(credentials: model.Credentials) -> None:
    payload = {"login": credentials.login, "password": credentials.password}

    _logger.debug(f"submitting post request to {common.status_url} with {payload=}")
    avoid_sls_crash()
    response = module_session.post(
        common.status_url, data=payload, timeout=CONNECT_TIMEOUT_SEC
    )

    msg = (
        f"posting payload {payload} to "
        f"{common.status_url} returns response {response}"
    )

    _logger.debug(msg)


def ports_from_range(_range: str | int) -> set[int]:
    """
    1770 -> {1770}
    "1770" -> {1770}
    "1770-1771" -> {1770, 1771}
    "1770-1770" -> {1770}
    "Request public session from broker" -> {}
    """

    u = str(_range)
    start = finish = u
    ports_in_range = set()

    if "-" in u:
        start, finish = [int(i) for i in u.split("-")]

    msg = (
        "returning early in ports_from_range() because"
        f" I can't cast '{start}' to int."
        " Not being able to parse this value is an"
        " expected condition since we're parsing a UI"
    )

    try:
        start = int(str(start).strip())
    except ValueError:
        _logger.debug(msg)
        return set()

    finish = int(str(finish).strip())
    ports = [start] if start == finish else list(range(start, finish + 1))

    for port in ports:
        ports_in_range.add(port)

    return ports_in_range


def get_remaining_unused_ports() -> list[int]:
    df_list = url_to_dataframe_list(common.status_url)
    df = get_session_port_map_dataframe(df_list)

    x = set([str(x) for x in df["ports"].tolist()])
    # x is mishmash of UI elements and port ranges and ports,
    # example: [1779, "1780-1782", 'Request public session from broker']
    _logger.debug(f"parsed ports from webui: {x}")

    ports_in_use = set()
    for _range in x:
        p = ports_from_range(_range)
        ports_in_use = ports_in_use | p

    _logger.debug(f"{sorted(ports_in_use)=}")

    y = set([int(x) for x in sls_listening_ports])
    _logger.debug(f"sls_listening_ports={sorted(y)}")

    remaining = y - ports_in_use

    msg = f"remaining ports availble for use {sorted(remaining)=}"
    _logger.debug(msg)

    remaining = sorted([int(x) for x in remaining])

    return remaining


def get_session_port_map_dataframe(
    df_list: typing.List[pandas.DataFrame],
) -> pandas.DataFrame:
    for df in df_list:
        if "enc" in df.columns and "dec" in df.columns:
            return df

    msg = "its unusual to not be able to find this port mapping"
    _logger.critical(msg)

    return pandas.DataFrame()


def get_incoming_ports_dataframe(
    df_list: typing.List[pandas.DataFrame],
) -> pandas.DataFrame:
    for df in df_list:
        if (
            "port" in df.columns
            and "encoders" in df.columns
            and "bitrate in" in df.columns
            and "bitrate out" in df.columns
        ):
            return df

    msg = "its unusual to not be able to find the Incoming Ports table"
    _logger.critical(msg)

    return pandas.DataFrame()


def dataframe_to_dict_list(df: pandas.DataFrame) -> typing.List[typing.Dict]:
    return df.to_dict("index")


def i_am_authenticated() -> bool:
    response = module_session.get(common.status_url)
    _logger.debug("\nauthenticated:")
    _logger.debug(response.text)
    if "sign in" in response.text.lower():
        return False
    return True


def url_to_dataframe_list(url: str) -> typing.List[pandas.DataFrame]:
    avoid_sls_crash()
    response = module_session.get(common.status_url)
    _logger.debug(response.text)
    df_list = html_to_dataframes(response.text)
    return df_list


def html_to_dataframes(html: str) -> typing.List[pandas.DataFrame]:
    _logger.debug(f"response from fetching {common.status_url}:")
    _logger.debug(html)

    try:
        df_list = pandas.read_html(html)
    except ValueError:
        _logger.warning("pandas.read_html caused exception")

    msg = f"There are {len(df_list):,} data frames in page {common.status_url}"
    _logger.debug(msg)

    for i, df in enumerate(df_list, 1):
        msg = f"data frame number {i}:"
        _logger.debug(msg)
        _logger.debug(df.head())

    return df_list


def html_to_text(html: str):
    soup = bs4.BeautifulSoup(html, "html.parser")
    text = soup.get_text().strip()
    text = text.strip()
    _logger.debug(f"beautiful soup parses html as text like this: {text}")

    return text


def post_sessioncreate_request(port: int, lifetime: datetime.timedelta) -> str:
    payload = {
        "port1": port,
        "port2": port,
        "lifetime": lifetime.total_seconds() / 3600,
        "request": "Request",
    }

    try:
        avoid_sls_crash()
        response = module_session.post(common.request_url, data=payload, timeout=5)
        _logger.debug(
            f"response from post request to {common.request_url} is {response.text}"
        )

        # Raises a HTTPError if the status is 4xx, 5xxx
        response.raise_for_status()

    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
        _logger.critical(f"can't reach {common.request_url}")
        sys.exit(-1)

    except requests.exceptions.HTTPError:
        _logger.critical("4xx, 5xx")
        sys.exit(-1)

    else:
        _logger.debug(f"Great!  I'm able to reach {common.request_url}")

    # refresh page so my new port appears on status page
    url = common.status_url.rstrip("?")
    url = f"{common.status_url}?"
    module_session.get(url)

    return response.text


def post_session_delete_request(session: model.LightSession) -> str:
    payload = {
        "action": 3,
        "idd1": session.encoder,
        "idd2": session.decoder,
    }

    avoid_sls_crash()
    response = module_session.post(common.request_url, data=payload, timeout=5)
    _logger.debug(
        f"response from post request to {common.request_url} is {response.text}"
    )

    return response.text


def generate_session_from_text(
    text: str, port: int, expires_at: datetime.timedelta
) -> model.LightSession:
    def debug(session):
        msg = (
            f"session {session} will expire at "
            f"{session.expires_at.astimezone(host_timezone)} utc: {session.expires_at}"
        )
        _logger.debug(msg)

    for line in text.splitlines():
        mo = session_pair_pat.search(line)
        if mo:
            dec = mo.group("dec").strip()
            enc = mo.group("enc").strip()
            session = model.LightSession(
                encoder=enc, decoder=dec, port=port, expires_at=expires_at
            )
            debug(session)
            return session

    session = model.LightSession(
        encoder="", decoder="", port=port, expires_at=expires_at
    )
    debug(session)

    return session


def check_sls_offline():
    avoid_sls_crash()
    response = module_session.get(common.status_url)
    _logger.debug(response.text)
    text = response.text
    msg = f"sls process isn't running according to parsing {common.status_url}"
    mo = sls_offline_pat.search(text)
    if mo:
        raise ValueError(msg)


def dataframe_list_to_list_of_lists_of_dicts(url: str) -> typing.List:
    df_list = url_to_dataframe_list(url)

    if df_list is None:
        msg = "data frame list is empty"
        _logger.critical(msg)
        return []

    df_list_as_list_of_list_of_dicts = []
    for df in df_list:
        _logger.debug("dataframe to dict list")
        dict_list = dataframe_to_dict_list(df)
        pf = pprint.pformat(dict_list)
        pf = f"\n{pf}"
        _logger.debug(pf)
        df_list_as_list_of_list_of_dicts.append(dict_list)

    return df_list_as_list_of_list_of_dicts


def show_list_of_dataframes_as_list_of_dicts():
    df_list = dataframe_list_to_list_of_lists_of_dicts(common.status_url)
    pf = pprint.pformat(df_list)
    pf = f"\n{pf}"
    _logger.debug("dataframe_list_to_list_of_lists_of_dicts")
    _logger.debug(pf)

    return df_list


def is_dataframe_empty(df: pandas.DataFrame):
    return df.empty


def display_dataframe(df: pandas.DataFrame):
    # display all the  rows
    pandas.set_option("display.max_rows", None)

    # display all the  columns
    pandas.set_option("display.max_columns", None)

    # set width  - 100
    pandas.set_option("display.width", 1000)

    # set column header -  left
    pandas.set_option("display.colheader_justify", "left")

    # set precision - 5
    pandas.set_option("display.precision", 5)

    _logger.debug(df)


def port_in_use(port: int) -> bool:
    df_list = url_to_dataframe_list(common.status_url)
    df = get_session_port_map_dataframe(df_list)
    display_dataframe(df)

    if is_dataframe_empty(df):
        _logger.warning("dataframe is empty")

    msg = f"{port} already exists, try another one please"
    if str(port) in df.ports.values:
        _logger.info(msg)
        return True
    return False


def get_data_usage_dataframe(
    df_list: typing.List[pandas.DataFrame],
) -> pandas.DataFrame | None:

    headers = {
        "ip",
        "data",
        "time",
        "first active",
        "last active",
    }

    for df in df_list:
        if headers.issubset(df.columns):
            return df

    msg = (
        "I searched through all tables"
        " but couldn't find the table whose field names matched"
        f" {headers}"
    )
    raise ValueError(msg)


def get_ip_addresses(column: str = "ip") -> list[str]:
    df_list = url_to_dataframe_list(common.status_url)
    df = get_data_usage_dataframe(df_list)
    _logger.debug(f"{column=}")
    ips = df[column].values
    _logger.debug(f"{ips=}")

    return ips


def set_global_sls_listening_ports():
    global sls_listening_ports
    df_list = url_to_dataframe_list(common.status_url)
    df = get_incoming_ports_dataframe(df_list)
    sls_listening_ports = list(set(df.port.values))


def get_random_port() -> int:
    _logger.debug(f"{sls_listening_ports=}")
    return random.choice(sls_listening_ports)


# WARNING THIS IS RACY since the port could be taken after checking
# whether its available
def get_available_port() -> int | None:
    remaining = get_remaining_unused_ports()
    if not remaining:
        return None
    return remaining[0]


def main(args):
    session_count = args.session_count
    session_lifetime_duration = durations.Duration(args.session_lifetime)

    d1 = args.prevent_crash_delay
    duration = durations.Duration(d1)
    delta = datetime.timedelta(seconds=duration.to_seconds())
    common.delay_to_prevent_crash = delta

    check_host_is_running(endpoint=common.status_url)

    login = os.getenv("VAR_SLS_LIGHT_WEBUI_LOGIN_USERNAME", None)
    password = os.getenv("VAR_SLS_LIGHT_WEBUI_LOGIN_PASSWORD", None)
    if not login:
        raise ValueError("VAR_SLS_LIGHT_WEBUI_LOGIN_USERNAME not defined")
    if not password:
        raise ValueError("VAR_SLS_LIGHT_WEBUI_LOGIN_PASSWORD not defined")
    creds = model.Credentials(login, password)

    populate_login_session(credentials=creds)

    if not i_am_authenticated():
        raise ValueError("not authenticated")

    session_lifetime = datetime.timedelta(
        seconds=session_lifetime_duration.to_seconds()
    )

    check_sls_offline()

    set_global_sls_listening_ports()

    ports_available = get_remaining_unused_ports()

    msg1 = (
        f"you asked for {session_count:,} ports, but there are only"
        f" {len(ports_available):,} remaining"
        f" available: {', '.join([str(x) for x in ports_available])}."
        " I will use the remaing ports"
    )

    if not ports_available:
        msg1 = (
            f"you asked for {session_count:,} port(s), but there are no"
            f" ports remaining available."
        )

    if len(ports_available) < session_count:
        _logger.warning(msg1)

    msg2 = f"you asked for {session_count:,} ports, but" " there are no ports availble"

    if get_available_port() is None:
        _logger.warning(msg2)
        _logger.warning("quitting prematurely")
        sys.exit(-1)

    if args.update_geolocation:
        get_remaining_unused_ports()
        # geolocation.ip_geolocation(get_ip_addresses())
        ips_list = get_ip_addresses()
        geolocate.get_regions_for_ips(ips_list)

    counter = session_count
    sessions = []
    while counter:
        port = get_available_port()
        _logger.debug(f"{port=}")

        if get_available_port() is None:
            msg = (
                f"you asked for {session_count:,} ports but I"
                f" could only create {len(sessions):,} ports"
                " before I ran out of available ports"
            )
            _logger.warning(msg)
            sys.exit(-1)

        html = post_sessioncreate_request(port=port, lifetime=session_lifetime)
        text = html_to_text(html)
        expires_at = (
            datetime.datetime.now(pytz.utc)
            + session_lifetime
            - datetime.timedelta(seconds=60)
        )
        session = generate_session_from_text(text, port=port, expires_at=expires_at)
        sessions.append(session)

        counter -= 1
        msg = (
            "session pair generated or re-fetched for "
            f"{port=} {session=}, {counter:,} remaining"
        )
        _logger.info(msg)

    for session in sessions:
        k8s.generate_k8s(session)
        scripts.generate_scripts(session)



import dataclasses
import logging

import jinja2

from . import common

_logger = logging.getLogger(__name__)

data_dir = common.data_dir / "scripts"


@dataclasses.dataclass
class ScriptContainer:
    script: str
    template: jinja2.Template


scripts = [
    ScriptContainer(
        script="sendstream.sh", template=common.env.get_template("sendstream.sh.j2")
    ),
    ScriptContainer(
        script="pulltest.sh", template=common.env.get_template("pulltest.sh.j2")
    ),
]


def generate_scripts(session):
    for container in scripts:
        generate_script(session, container)


def generate_script(session, container):
    data = common.generate_data(session)

    x1 = str(data["pull_port"])
    y1 = session.encoder.replace("$", "")
    z1 = f"{x1}-{y1}"

    _dir = data_dir / z1
    path = _dir / container.script
    text = container.template.render(data=data)
    _logger.info(f"writing {path.resolve()}")

    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(text)
